# gisweb_tenants/crypto.py
from __future__ import annotations
import base64, secrets
from dataclasses import dataclass
from typing import Optional, Iterable
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from cryptography.exceptions import InvalidTag

def _b64url_decode(s: str) -> bytes:
    pad = "=" * ((4 - len(s) % 4) % 4)
    return base64.urlsafe_b64decode(s + pad)

def _is_b64url(s: str) -> bool:
    return all(c.isalnum() or c in "-_" for c in s)

def coerce_key_to_bytes(key: str | bytes) -> bytes:
    
    if isinstance(key, bytes):
        return key
    k = key.strip()
    # 1) HEX
    try:
        if all(c in "0123456789abcdefABCDEF" for c in k) and len(k) % 2 == 0:
            return bytes.fromhex(k)
    except Exception:
        pass
    # 2) base64 URL-safe (con o senza "=")
    try:
        pad = "=" * ((4 - len(k) % 4) % 4)
        return base64.urlsafe_b64decode(k + pad)
    except Exception:
        pass
    # 3) base64 standard
    try:
        pad = "=" * ((4 - len(k) % 4) % 4)
        import base64 as b64
        return b64.b64decode(k + pad)
    except Exception:
        pass
    raise ValueError("Chiave AEAD non è hex né base64 valida")

@dataclass(frozen=True)
class AeadBox:
    key: bytes  # 16|24|32

    @staticmethod
    def from_text(key: str | bytes) -> "AeadBox":
        raw = coerce_key_to_bytes(key)
        if len(raw) not in (16, 24, 32):
            raise ValueError("AES key must be 16/24/32 bytes")
        return AeadBox(raw)

    def encrypt(self, plaintext: bytes, aad: Optional[bytes] = None) -> bytes:
        nonce = secrets.token_bytes(12)
        ct = AESGCM(self.key).encrypt(nonce, plaintext, aad)
        return nonce + ct

    def decrypt(self, payload: bytes, aad: Optional[bytes] = None) -> bytes:
        if len(payload) < 13:
            raise ValueError("ciphertext too short")
        nonce, ct = payload[:12], payload[12:]
        return AESGCM(self.key).decrypt(nonce, ct, aad)

# ---- supporto registro {$enc:aesgcm, v:1, n:..., ct:..., [aad: "..."]}

def decrypt_field(
    value: str | dict,
    box: AeadBox,
    *,
    aad: Optional[bytes] = None,
    field: Optional[str] = None,
    tenant: Optional[str] = None,
    try_common: bool = True,
) -> str:
    """
    Decifra un campo del registry.
    - Se 'value' è stringa, la restituisce.
    - Se è dict {$enc:'aesgcm', v, n, ct, [aad]}, usa AAD esplicita oppure tenta
      una lista di AAD candidate (field, v, tenant) per compat con versioni precedenti.
    """
    if isinstance(value, str):
        return value

    if not (isinstance(value, dict) and value.get("$enc") == "aesgcm"):
        raise ValueError("Formato campo non supportato per decrypt_field")

    v = value.get("v")
    nonce = _b64url_decode(value["n"])
    ct = _b64url_decode(value["ct"])
    payload = nonce + ct

    # 1) se l'aad è nel dict, usala subito
    if "aad" in value and value["aad"] is not None:
        cand = value["aad"]
        aad_bytes = cand.encode("utf-8") if isinstance(cand, str) else cand
        return box.decrypt(payload, aad=aad_bytes).decode("utf-8")

    # 2) costruisci lista di candidate
    candidates: list[Optional[bytes]] = []
    if aad is not None:
        candidates.append(aad)
    if try_common:
        if field:
            candidates += [field.encode("utf-8")]
        if isinstance(v, int):
            candidates += [f"v={v}".encode("utf-8"), f"aesgcm|v={v}".encode("utf-8")]
        if tenant and field and isinstance(v, int):
            candidates += [f"{tenant}|{field}|v={v}".encode("utf-8")]
        # ultima spiaggia: nessuna AAD
        candidates.append(None)

    last_err: Optional[Exception] = None
    for cand in candidates:
        try:
            return box.decrypt(payload, aad=cand).decode("utf-8")
        except InvalidTag as e:
            last_err = e
            continue

    raise InvalidTag(f"Impossibile decifrare: chiave o AAD non corretti. Tentativi: {len(candidates)}") from last_err


def encrypt_field(plaintext: str, box: AeadBox, *, aad: bytes | None = None, version: int = 1) -> dict:
    payload = box.encrypt(plaintext.encode("utf-8"), aad=aad)
    n, ct = payload[:12], payload[12:]
    return {"$enc": "aesgcm", "v": version,
            "n": base64.urlsafe_b64encode(n).decode().rstrip("="),
            "ct": base64.urlsafe_b64encode(ct).decode().rstrip("="),
            "aad": aad.decode() if isinstance(aad, (bytes, bytearray)) else aad}




# gisweb_tenants/registry.py

from __future__ import annotations
from dataclasses import dataclass
from functools import lru_cache
from pathlib import Path
from typing import Dict, Any, Tuple, Optional

import hashlib
import yaml
from sqlalchemy.engine import make_url, URL

from .crypto import AeadBox, decrypt_field 

# version -> (entries, aead_key_bytes)
_REG_SNAPSHOTS: Dict[int, tuple[dict[str, "TenantRecord"], Optional[bytes]]] = {}
_MAX_SNAPSHOTS = 8

@lru_cache(maxsize=512)
def _resolve_triplet_cached(version: int, tenant: str) -> Tuple[str, str, str]:
    entries, aead_key = _REG_SNAPSHOTS[version]
    rec = entries.get(tenant)
    if rec is None:
        raise KeyError(f"Tenant '{tenant}' non trovato")
    cfg = rec.config or {}

    db_name = cfg.get("db_name") or tenant
    db_user = cfg.get("db_user")
    db_password = cfg.get("db_password")
        
    # decrypt se necessario: AAD fissa come da tuo encrypt.py
    if isinstance(db_user, dict) and db_user.get("$enc") == "aesgcm":
        if not aead_key:
            raise RuntimeError("Credenziali cifrate ma manca chiave AEAD")
        box = AeadBox.from_text(aead_key)
        db_user = decrypt_field(db_user, box, aad=f"{tenant}|db|user".encode())

    if isinstance(db_password, dict) and db_password.get("$enc") == "aesgcm":
        if not aead_key:
            raise RuntimeError("Credenziali cifrate ma manca chiave AEAD")
        box = AeadBox.from_text(aead_key)
        db_password = decrypt_field(db_password, box, aad=f"{tenant}|db|password".encode())

    if not db_user or not db_password:
        raise RuntimeError(f"Credenziali DB mancanti per tenant '{tenant}'")

    return db_name, db_user, db_password


@dataclass(frozen=True)
class TenantRecord:
    name: str
    config: Dict[str, Any]


class TenantsRegistry:
    """
    Registry con cache LRU per triplet decrittati.
    - Se istanzi da file, invalida automaticamente quando cambia l'mtime.
    - Se istanzi da testo, la "versione" deriva da hash(text + aead_key).
    """

    def __init__(self, *, text: Optional[str] = None, path: Optional[Path] = None, aead_key: Optional[str | bytes] = None):
        if not text and not path:
            raise ValueError("Serve 'text' YAML oppure 'path' al file del registry")
        self._text = text
        self._path = Path(path) if path else None
        self._aead_key: Optional[bytes] = (aead_key.encode() if isinstance(aead_key, str) else aead_key) if aead_key else None
        self._version: Optional[int] = None  # int(mtime) oppure hash del testo
        self._ensure_loaded()

    # --------- API pubblico

    def exists(self, tenant: str) -> bool:
        self._ensure_loaded()
        entries, _ = _REG_SNAPSHOTS[self._version]  # type: ignore[arg-type]
        return tenant.strip().lower() in entries

    def get(self, tenant: str) -> TenantRecord:
        self._ensure_loaded()
        entries, _ = _REG_SNAPSHOTS[self._version]  # type: ignore[arg-type]
        rec = entries.get(tenant.strip().lower())
        if not rec:
            raise KeyError(f"Tenant '{tenant}' non trovato")
        return rec

    def resolve_triplet(self, tenant: str) -> Tuple[str, str, str]:
        self._ensure_loaded()
        return _resolve_triplet_cached(self._version, tenant.strip().lower())  # type: ignore[arg-type]

    def build_dsn(self, base_dsn: str, tenant: str) -> str:
        """
        Costruisce un DSN per il tenant a partire da un DSN base (driver/host/port).
        Esempio base_dsn: "postgresql+asyncpg://@localhost:6432/postgres"
        """
        db_name, db_user, db_password = self.resolve_triplet(tenant)
        base: URL = make_url(base_dsn)
        url = base.set(database=db_name, username=db_user, password=db_password)
        return url

    @property
    def names(self) -> list[str]:
        self._ensure_loaded()
        entries, _ = _REG_SNAPSHOTS[self._version]  # type: ignore[arg-type]
        return list(entries.keys())

    def invalidate(self) -> None:
        """Forza ricarica alla prossima chiamata."""
        self._version = None

    # --------- privato

    def _ensure_loaded(self) -> None:
        # Se da file: versione = int(mtime). Se da testo: hash(text + aead_key)
        if self._path:
            try:
                mtime = int(self._path.stat().st_mtime)
            except FileNotFoundError:
                raise RuntimeError(f"Registry file non trovato: {self._path}")
            if self._version != mtime:
                text = self._path.read_text(encoding="utf-8")
                self._load_into_snapshots(text, version=mtime)
                self._version = mtime
        else:
            # versione derivata da hash
            text = self._text or ""
            h = hashlib.sha1()
            h.update(text.encode("utf-8"))
            if self._aead_key:
                h.update(self._aead_key)
            ver = int.from_bytes(h.digest()[:4], "big")  # un int stabile a 32 bit
            if self._version != ver:
                self._load_into_snapshots(text, version=ver)
                self._version = ver

    def _load_into_snapshots(self, yaml_text: str, *, version: int) -> None:
        data = yaml.safe_load(yaml_text) or {}
        tenants = data.get("tenants") or {}
        entries: dict[str, TenantRecord] = {}
        for name, cfg in tenants.items():
            k = str(name).strip().lower()
            entries[k] = TenantRecord(name=k, config=cfg or {})
        _REG_SNAPSHOTS[version] = (entries, self._aead_key)
        while len(_REG_SNAPSHOTS) > _MAX_SNAPSHOTS:
            _REG_SNAPSHOTS.pop(next(iter(_REG_SNAPSHOTS)))


from fastapi import HTTPException, status

class TenantNotFound(HTTPException):
    def __init__(self, tenant: str):
        super().__init__(status_code=status.HTTP_404_NOT_FOUND,
                         detail=f"Tenant '{tenant}' non trovato nel registry")

class TenantForbidden(HTTPException):
    def __init__(self):
        super().__init__(status_code=status.HTTP_403_FORBIDDEN,
                         detail="Tenant non autorizzato")

class SessionStoreUnavailable(HTTPException):
    def __init__(self):
        super().__init__(status_code=503, detail="Session store non disponibile")



# gisweb_tenants/config.py
from __future__ import annotations
from dataclasses import dataclass
from typing import Literal
from sqlalchemy.engine import make_url
from pathlib import Path

Mode = Literal["development", "testing", "production"]

@dataclass(frozen=True)
class TenantSettings:
    # Ambiente
    MODE: Mode = "development"

    # Multitenant HTTP
    TENANT_HEADER: str = "X-Tenant"
    DEFAULT_TENANT: str = "istanze"

    # DB di base (host, porta, driver, utente/password di default)
    # Il nome database verrà sovrascritto per-tenant dal registry quando presente.
    ASYNC_DATABASE_URI: str = "postgresql+asyncpg://postgres:postgres@localhost:6432/postgres"

    # SQLAlchemy
    ECHO_SQL: bool = False
    POOL_SIZE: int = 10

    # Verrà usato come application_name: "<APP_NAME_PREFIX>:<tenant>"
    APP_NAME_PREFIX: str = "fastapi"

    def drivername(self) -> str:
        """
        Ritorna lo scheme del driver, es.:
        - 'postgresql+asyncpg'
        - 'postgresql+psycopg'
        Serve a scegliere dove mettere application_name (server_settings vs connect_args).
        """
        return make_url(self.ASYNC_DATABASE_URI).drivername


@dataclass(frozen=True, slots=True)
class TenantsConfig:
    tenants_file: Path                 # path a tenants.yml
    tenant_header: str = "X-Tenant"
    default_tenant: str = "istanze"
    allowed_tenants_csv: str = ""      # "a,b,c"
    strict_whitelist: bool = False     # se True -> 403 se non in allowed

@dataclass(frozen=True, slots=True)
class CryptoConfig:
    encrypt_key: bytes  # 32 bytes
    
@dataclass(frozen=True, slots=True)
class DbDefaults:
    scheme: str = "postgresql+asyncpg"
    host: str = "localhost"
    port: int = 6432


# gisweb_tenants/db.py

from __future__ import annotations
from typing import Dict, AsyncIterator
from contextlib import asynccontextmanager

from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine, async_sessionmaker, AsyncSession
from sqlalchemy.engine import make_url
from sqlalchemy.pool import NullPool

from .config import TenantSettings
from .registry import TenantsRegistry, TenantRecord

_engine_registry: Dict[str, AsyncEngine] = {}

def _build_connect_args(settings: TenantSettings, tenant: str) -> dict:
    app_name = f"{settings.APP_NAME_PREFIX}:{tenant}"
    driver = settings.drivername()
    if "+asyncpg" in driver:
        return {"server_settings": {"application_name": app_name}}
    if "+psycopg" in driver:
        return {"options": f"-c application_name={app_name}"}
    return {"application_name": app_name}

def _engine_args(settings: TenantSettings) -> dict:
    testing = settings.MODE == "testing"
    return {
        "echo": settings.ECHO_SQL,
        "pool_pre_ping": True,
        "pool_size": None if testing else settings.POOL_SIZE,
        "max_overflow": 64 if not testing else 0,
        "poolclass": NullPool if testing else None,
    }

def get_engine(settings: TenantSettings, tenant: str, registry: TenantsRegistry) -> AsyncEngine:
    
    key = f"{tenant}"
    eng = _engine_registry.get(key)
    if eng is not None:
        return eng

    url = registry.build_dsn(settings.ASYNC_DATABASE_URI, tenant)
    print (str(url))
    connect_args = _build_connect_args(settings, tenant)
    args = {k: v for k, v in _engine_args(settings).items() if v is not None}
    eng = create_async_engine(url, connect_args=connect_args, **args)
    _engine_registry[key] = eng
    return eng

def get_sessionmaker(settings: TenantSettings, tenant: str, registry: TenantsRegistry) -> async_sessionmaker[AsyncSession]:
    engine = get_engine(settings, tenant, registry)
    return async_sessionmaker(engine, expire_on_commit=False)


@asynccontextmanager
async def tenant_session(settings: TenantSettings, tenant: str, registry: TenantsRegistry | None = None) -> AsyncIterator[AsyncSession]:
    sm = get_sessionmaker(settings, tenant, registry)
    async with sm() as session:
        yield session


# ------------------------------
# 🔻 gestione shutdown/cleanup
# ------------------------------

async def dispose_engine(tenant: str) -> bool:
    """
    Chiude e rimuove l'engine del tenant (se esiste).
    Ritorna True se c'era qualcosa da chiudere.
    """
    eng = _engine_registry.pop(tenant, None)
    if not eng:
        return False
    try:
        await eng.dispose()
    except Exception:
        # non alziamo in shutdown
        return False
    return True

async def dispose_all_engines() -> dict:
    """
    Chiude tutti gli AsyncEngine creati da questo modulo e svuota la registry.
    Idempotente e sicuro da chiamare più volte.
    Ritorna un report {"closed": [...], "failed": [...]}.
    """
    closed, failed = [], []
    # copia chiavi per evitare modifiche durante l'iterazione
    keys = list(_engine_registry.keys())
    for tenant in keys:
        eng = _engine_registry.pop(tenant, None)
        if not eng:
            continue
        try:
            await eng.dispose()
            closed.append(tenant)
        except Exception:
            failed.append(tenant)
    return {"closed": closed, "failed": failed}